# cython: language_level=3
import contextlib
import os
import pkg_resources

from libc.stdlib cimport free, malloc
from libc.string cimport strcpy


@contextlib.contextmanager
def suppress_stdout():
    null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
    # Save the actual stdout (1) and stderr (2) file descriptors.
    save_fds = [os.dup(1), os.dup(2)]

    os.dup2(null_fds[0], 1)
    os.dup2(null_fds[1], 2)

    yield

    # Re-assign the real stdout/stderr back to (1) and (2)
    os.dup2(save_fds[0], 1)
    os.dup2(save_fds[1], 2)
    # Close the null files
    for fd in null_fds + save_fds:
        os.close(fd)


cdef extern from "udunits2.h":
    ctypedef struct ut_system:
        pass
    ctypedef struct ut_unit:
        pass

    ut_system* ut_read_xml(const char * path)
    void ut_free_system(ut_system* system)
    ut_unit* ut_parse(const ut_system* system, const char* string, int encoding)
    int ut_are_convertible(const ut_unit* unit1, const ut_unit* unit2)
    int ut_format(const ut_unit* unit, char* buf, size_t size, unsigned opts)
    int ut_is_dimensionless(const ut_unit* unit)


cdef class Units:
    cdef ut_system* _c_system
    cdef char[2048] STR_BUFFER
    cdef char* _filepath

    def __cinit__(self):
        try:
            filepath = os.environ["UDUNITS2_XML_PATH"]
        except KeyError:
            filepath = pkg_resources.resource_filename(
                "bmi_tester", "data/udunits/udunits2.xml"
            )
        as_bytes = str(filepath).encode("utf-8")
        self._filepath = <char*>malloc((len(as_bytes) + 1) * sizeof(char))
        strcpy(self._filepath, as_bytes)

        with suppress_stdout():
            self._c_system = ut_read_xml(self._filepath)

        # self._c_system = ut_read_xml(NULL)

    def norm(self, units):
        unit = ut_parse(self._c_system, units.encode("utf-8"), 0)
        str_len = ut_format(unit, self.STR_BUFFER, 2048, 4 | 8)
        return self.STR_BUFFER

    def is_valid(self, units):
        return ut_parse(self._c_system, units.encode("utf-8"), 0) != NULL

    def is_time(self, units):
        src_unit = ut_parse(self._c_system, units.encode("utf-8"), 0)
        dst_unit = ut_parse(self._c_system, "s".encode("utf-8"), 0)
        return bool(ut_are_convertible(src_unit, dst_unit))

    def is_dimensionless(self, units):
        unit = ut_parse(self._c_system, units.encode("utf-8"), 0)
        return ut_is_dimensionless(unit) != 0

    def __dealloc__(self):
        ut_free_system(self._c_system)
        free(self._filepath)
        self._c_system = NULL
        self._filepath = NULL



import numpy as np
import pytest

from ..conftest import skip_if_grid_type_is


# @pytest.mark.dependency(depends=["test_get_grid_type", "test_get_grid_rank"], scope="session")
def test_grid_x(initialized_bmi, gid):
    skip_if_grid_type_is(initialized_bmi, gid, "uniform_rectilinear")

    gtype = initialized_bmi.get_grid_type(gid)
    ndim = initialized_bmi.get_grid_rank(gid)
    if ndim < 1:
        pytest.skip(f"grid is rank {ndim}")

    if gtype == "unstructured":
        size = initialized_bmi.get_grid_node_count(gid)
    elif gtype == "rectilinear":
        shape = np.empty(ndim, dtype=np.int32)
        initialized_bmi.get_grid_shape(gid, shape)
        size = shape[-1]

    x = np.full(size, np.nan)
    assert x is initialized_bmi.get_grid_x(gid, x)
    assert np.all(~np.isnan(x))


# @pytest.mark.dependency(depends=["test_get_grid_type", "test_get_grid_rank"], scope="session")
def test_grid_y(initialized_bmi, gid):
    skip_if_grid_type_is(initialized_bmi, gid, "uniform_rectilinear")

    gtype = initialized_bmi.get_grid_type(gid)
    ndim = initialized_bmi.get_grid_rank(gid)
    if ndim < 2:
        pytest.skip(f"grid is rank {ndim}")

    if gtype == "unstructured":
        size = initialized_bmi.get_grid_node_count(gid)
    elif gtype == "rectilinear":
        shape = np.empty(ndim, dtype=np.int32)
        initialized_bmi.get_grid_shape(gid, shape)
        size = shape[-2]

    y = np.full(size, np.nan)
    assert y is initialized_bmi.get_grid_y(gid, y)
    assert np.all(~np.isnan(y))


# @pytest.mark.dependency(depends=["test_get_grid_type", "test_get_grid_rank"], scope="session")
def test_grid_z(initialized_bmi, gid):
    skip_if_grid_type_is(initialized_bmi, gid, "uniform_rectilinear")

    gtype = initialized_bmi.get_grid_type(gid)
    ndim = initialized_bmi.get_grid_rank(gid)
    if ndim < 3:
        pytest.skip(f"grid is rank {ndim}")

    if gtype == "unstructured":
        size = initialized_bmi.get_grid_node_count(gid)
    elif gtype == "rectilinear":
        shape = np.empty(ndim, dtype=np.int32)
        initialized_bmi.get_grid_shape(gid, shape)
        size = shape[-3]

    z = np.full(size, np.nan)
    assert z is initialized_bmi.get_grid_z(gid, z)
    assert np.all(~np.isnan(z))



import warnings

import numpy as np

from ..conftest import skip_if_grid_type_is_not


# @pytest.mark.dependency(depends=["test_get_grid_rank"], scope="session")
def test_get_grid_shape(initialized_bmi, gid):
    """Test the grid shape."""
    skip_if_grid_type_is_not(
        initialized_bmi,
        gid,
        ("uniform_rectilinear", "rectilinear", "structured_quadrilateral"),
    )

    ndim = initialized_bmi.get_grid_rank(gid)

    shape = np.empty(ndim, dtype=np.int32)
    try:
        rtn = initialized_bmi.get_grid_shape(gid, shape)
    except TypeError:
        warnings.warn("get_grid_shape should take two arguments")
        rtn = initialized_bmi.get_grid_shape(gid)
        shape[:] = rtn
    else:
        assert rtn is shape

    assert np.all(shape > 0)


# @pytest.mark.dependency(depends=["test_get_grid_rank"], scope="session")
def test_get_grid_spacing(initialized_bmi, gid):
    """Test the grid spacing."""
    skip_if_grid_type_is_not(initialized_bmi, gid, "uniform_rectilinear")

    ndim = initialized_bmi.get_grid_rank(gid)

    spacing = np.empty(ndim, dtype=float)
    assert spacing is initialized_bmi.get_grid_spacing(gid, spacing)
    assert np.all(spacing > 0.0)


# @pytest.mark.dependency(depends=["test_get_grid_rank"], scope="session")
def test_get_grid_origin(initialized_bmi, gid):
    """Test the grid origin."""
    skip_if_grid_type_is_not(initialized_bmi, gid, "uniform_rectilinear")

    ndim = initialized_bmi.get_grid_rank(gid)

    spacing = np.empty(ndim, dtype=float)
    assert spacing is initialized_bmi.get_grid_spacing(gid, spacing)
    assert np.all(spacing > 0.0)



import numpy as np


class Bmi(object):
    _input_var_names = ("land_surface__elevation", "land_surface_air__temperature")
    _output_var_names = (
        "land_surface__elevation",
        "land_surface~10m-above_air_flow__speed",
    )

    _var_units = {
        "land_surface__elevation": "m",
        "land_surface_air__temperature": "C",
        "land_surface~10m-above_air_flow__speed": "m/s",
    }

    def __init__(self):
        self._values = {
            "land_surface__elevation": np.empty(12, dtype=float),
            "land_surface_air__temperature": np.empty(12, dtype=float),
            "land_surface~10m-above_air_flow__speed": np.empty(12, dtype=float),
        }

    def initialize(self, config_file):
        with open(config_file, "r"):
            pass

    def update(self):
        pass

    def finalize(self):
        pass

    def get_component_name(self):
        return "Test BMI component"

    def get_input_var_names(self):
        return self._input_var_names

    def get_output_var_names(self):
        return self._output_var_names

    def get_current_time(self):
        return 0.0

    def get_start_time(self):
        return 0.0

    def get_end_time(self):
        return 1.0

    def get_time_step(self):
        return 1.0

    def get_time_units(self):
        return "s"

    def get_var_grid(self, name):
        return 0

    def get_var_itemsize(self, name):
        return 8

    def get_var_nbytes(self, name):
        return 8 * 12

    def get_var_type(self, name):
        return "float"

    def get_var_units(self, name):
        return self._var_units[name]

    def get_var_location(self, name):
        return "node"

    def get_grid_rank(self, id_):
        return 2

    def get_grid_size(self, id_):
        return 12

    def get_grid_node_count(self, id_):
        return 12

    def get_grid_type(self, id_):
        return "uniform_rectilinear"

    def get_grid_shape(self, id_, shape):
        shape[:] = (3, 4)
        return shape

    def get_grid_spacing(self, id_, spacing):
        spacing[:] = (10.0, 20.0)
        return spacing

    def get_value(self, name, buffer):
        buffer[:] = self._values[name]
        return buffer

    def set_value(self, name, buffer):
        self._values[name][:] = buffer
        return buffer
